from os import cpu_count
import click as click

import openvariant
from openvariant.tasks.cat import cat as cat_task
from openvariant.tasks.count import count as count_task
from openvariant.tasks.groupby import group_by as group_by_task
from openvariant.tasks.plugin import PluginActions


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(openvariant.__version__)
def openvar():
    """'openvar' is the command-line interface of OpenVariant.
    Parsing and data transformation of multiple input formats."""
    pass


@openvar.command(name="cat", short_help='Concatenate parsed files to standard output.')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--where', '-w', type=click.STRING, default=None, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--annotations', '-a', type=click.Path(exists=True), default=None,
              help="Annotation path. eg: /path/annotation_vcf.yaml")
@click.option('--header', is_flag=True, help="Show the result header.")
@click.option('--output', '-o', default=None, help="File to write the output.")
def cat(input_path: str, where: str or None, annotations: str or None, header: bool, output: str or None):
    """Print the parsed files on the stdout/"output"."""
    cat_task(input_path, annotations, where, header, output)


@openvar.command(name="count", short_help='Number of rows that matches a specified criterion.')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--where', '-w', multiple=False, type=click.STRING, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--group_by', '-g', type=click.STRING, help="Key to group rows. eg: COUNTRY")
@click.option('--annotations', '-a', default=None, type=click.Path(exists=True),
              help="Annotation path. eg: /path/annotation_vcf.yaml")
@click.option('--cores', '-c', type=click.INT, default=cpu_count(), help='Maximum processes to run in parallel.')
@click.option('--quite', '-q', is_flag=True, help="Don't show the progress.")
@click.option('--output', '-o', default=None, help="File to write the output.")
def count(input_path: str, where: str, group_by: str, cores: int, quite: bool, annotations: str or None,
          output: str or None) -> None:
    """Print on the stdout/"output" the number of rows that meets the criteria."""
    result = count_task(input_path, annotations, group_by=group_by, where=where, cores=cores, quite=quite)
    out_file = None
    if output:
        out_file = open(output, "w")
    if len(result[1]) > 0:
        for k, v in sorted(result[1].items(), key=lambda res: res[1]):
            if output:
                out_file.write("{}\t{}\n".format(k, v))
            else:
                print("{}\t{}".format(k, v))

    if output:
        out_file.write("TOTAL\t{}\n".format(result[0]))
    else:
        print("TOTAL\t{}".format(result[0]))

    if output:
        out_file.close()


@openvar.command(name="groupby", short_help='Group the parsed result for each different value of the specified key.')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--header', is_flag=True, help="Show the result header.")
@click.option('--show', is_flag=True, help='Show group by each row.')
@click.option('--where', '-w', type=click.STRING, default=None, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--group_by', '-g', type=click.STRING, default=None, help="Key to group rows. eg: COUNTRY")
@click.option('--script', '-s', type=click.STRING, default=None,
              help="Filter expression. eg: gzip > \${GROUP_KEY}.parsed.tsv.gz")
@click.option('--annotations', '-a', default=None, type=click.Path(exists=True),
              help="Annotation path. eg: /path/annotation_vcf.yaml")
@click.option('--cores', '-c', type=click.INT, default=cpu_count(), help='Maximum processes to run in parallel.')
@click.option('--quite', '-q', is_flag=True, help="Don't show the progress.")
@click.option('--output', '-o', help="File to write the output.", default=None)
def groupby(input_path: str, script: str, where: str, group_by: str, cores: int, quite: bool, annotations: str or None,
            header: bool, show: bool, output: str or None):
    """Print on the stdout/"output" the parsed files group by a specified field."""
    out_file = None
    if output:
        out_file = open(output, 'w')
    for group_key, group_result, command in group_by_task(input_path, annotations, script, key_by=group_by, where=where,
                                                          cores=cores, quite=quite, header=header):
        for r in group_result:
            if command:
                if output:
                    out_file.write(f"{group_key}\t{r}\n") if show else out_file.write(f"{r}\n")
                else:
                    print(f"{group_key}\t{r}") if show else print(f"{r}")
            else:
                if header:
                    if output:
                        out_file.write(f"{r}\n")
                    else:
                        print(f"{r}")
                    header = False
                else:
                    if output:
                        out_file.write(f"{group_key}\t{r}\n") if show else out_file.write(f"{r}\n")
                    else:
                        print(f"{group_key}\t{r}") if show else print(f"{r}")
    if output:
        out_file.close()


@openvar.command(name="plugin", short_help='Actions to execute for a plugin: create.')
@click.argument('action', type=click.Choice(['create']))
@click.option('--name', '-n', type=click.STRING, help="Name of the plugin.")
@click.option('--directory', '-d', type=click.STRING, help="Directory to reach the plugin.")
def plugin(action, name: str or None, directory: str or None):
    """Actions to apply on the plugin system."""
    PluginActions[action.upper()].value(name, directory)


if __name__ == "__main__":
    openvar()



"""
Find files
====================================
A core functionally to search files and with its corresponding annotation.
"""
import glob
from os import listdir
from os.path import isfile, join, isdir, dirname
from typing import Generator

from openvariant.annotation.annotation import Annotation
from openvariant.annotation.config_annotation import ANNOTATION_EXTENSION
from openvariant.utils.utils import check_extension


def _get_annotation(file_path, annotation):
    if annotation is not None:
        try:
            for ext, _ in annotation.structure.items():
                if check_extension(ext, file_path):
                    yield file_path, annotation
        except AttributeError:
            raise AttributeError("Unable to parse annotation file, check its location.")


def _scan_files(base_path: str, annotation: Annotation, fix: bool):
    """Recursive exploration from a base path"""
    if isdir(base_path):
        if not fix:
            for annotation_file in glob.iglob(join(base_path, "*.{}".format(ANNOTATION_EXTENSION))):
                annotation = Annotation(annotation_file)
        list_files = listdir(base_path)
        list_files.sort()
        for file_name in list_files:
            file_path = join(base_path, file_name)
            try:
                for f, a in _scan_files(file_path, annotation, fix):
                    yield f, a
            except PermissionError as e:
                raise PermissionError(f"Unable to open {file_name}: {e}")
    elif isfile(base_path):
        file_path = base_path
        try:
            for f, a in _get_annotation(file_path, annotation):
                yield f, a
        except PermissionError as e:
            raise PermissionError(f"Unable to open {base_path}: {e}")
    else:
        raise FileNotFoundError(f"Unable to open {base_path}, it's not a file nor a directory.")


def _find_files(base_path: str, annotation: Annotation or None, fix: bool) -> Generator[str, Annotation, None]:
    """Recursive exploration from a base path distinct if there's a fix annotation or no"""
    if not fix:
        if isfile(base_path):
            annotation_path = dirname(base_path)
        else:
            annotation_path = base_path
        for annotation_file in glob.iglob(join(annotation_path, "*.{}".format(ANNOTATION_EXTENSION))):
            annotation = Annotation(annotation_file)
            
    for f, a in _scan_files(base_path, annotation, fix):
        yield f, a


def findfiles(base_path: str, annotation_path: str or None = None) -> Generator[str, Annotation, None]:
    """Get each file and its proper annotation object.

    Parameters
    ----------
    base_path : srt
        Base path of input folder/file.
    annotation_path : str or None
        Path of annotation file.

    Yields
    -------
    str
        Input file's name.
    Annotation
        The proper schema of each input file.
    """
    annotation, fix = (Annotation(annotation_path), True) if annotation_path is not None else (None, False)
    for f, a in _find_files(base_path, annotation, fix):
        yield f, a


import pkg_resources

from openvariant.annotation.annotation import Annotation
from openvariant.tasks import cat, count, group_by
from openvariant.variant import Variant
from openvariant.find_files import findfiles

version = pkg_resources.require("open-variant")[0].version
__version__ = version

__all__ = ['Annotation', 'Variant', 'cat', 'count', 'group_by', 'findfiles']




import re
from fnmatch import fnmatch
from os.path import basename


def check_extension(ext: str, path: str) -> bool:
    """Check if file matches with the annotation pattern"""
    return fnmatch(basename(path), ext) if ext[0] == '*' else re.match(ext, basename(path)) is not None



from .context import Context
from .plugin import Plugin
from .get_AF import Get_afPlugin, Get_afContext
from .alteration_type import Alteration_typePlugin, Alteration_typeContext

__all__ = ['Plugin', 'Context', 'Get_afPlugin', 'Get_afContext', 'Alteration_typePlugin', 'Alteration_typeContext']



from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="open-variant",
    version="0.7.6",
    author="BBGLab - Barcelona Biomedical Genomics Lab",
    author_email='bbglab@irbbarcelona.org',
    description="OpenVariant provides different functionalities to read, parse and operate different multiple input "
                "file formats, being able to customize the output.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD 3-Clause License',
    keywords='bioinformatics,openvariant,openvar,bbglab',
    packages=find_packages(exclude=["tests.*", "tests"]),
    include_package_data=True,
    install_requires=['pyyaml', 'tqdm', 'click'],
    entry_points={
        'console_scripts': [
            'openvar = openvariant.commands.openvar:openvar',
        ]
    },
    url="https://github.com/bbglab/openvariant",
)
